'use strict';

const AWS = require('aws-sdk');
//const inflection = require('inflection');

// BEGIN : To be removed from lambda deployment
AWS.config.update({region:'us-east-1'});
// END

const dynamo = new AWS.DynamoDB.DocumentClient();
const tableName = 'JsonApiTable';
const EPOCH = 1545907609050;
const MAX_OBJECTS=1000; // Max number of objects returned by finder


// Instagram inspired --> https://instagram-engineering.com/sharding-ids-at-instagram-1cf5a71e5a5c
function generateRowId(subid) {
    var ts = new Date().getTime() - EPOCH; // limit to recent
    // 41 bits for time in milliseconds (gives us 41 years of IDs with a custom epoch
    var randid = Math.floor(Math.random() * 512);
    ts = (ts * 64);   // bit-shift << 6
    // Given shard (if any...)
    ts = ts + subid;
    // random value
    return (ts * 512) + (randid % 512);
  }
  
  const createObject = (obj) => {
      let objout = {  "type": obj.ObjectType,
                      "id": obj.Id,
                      "attributes": {},
                      "relationships": {}
                  };
      for (var attr in obj) {
          if (attr !== 'ObjectType' && attr !== 'Id') {
              if(!attr.endsWith('_id')) {
                  objout.attributes[attr] = obj[attr]; 
              } else {
                  let relationDetails = attr.split('_');
                  if (obj[attr].startsWith('[')) {
                    // hasMany relationship
                    let idsArray = obj[attr].substring(1, obj[attr].length-1).split(',');
                    let relData = [];
                    for (let l=0;l<idsArray.length;l++) {
                        let relObject = {"type": relationDetails[0], "id": Number(idsArray[l]) };
                        relData.push(relObject);                    }
                    objout.relationships[relationDetails[0]] = {
                        "links": {
                            "self": "/"+obj.ObjectType+"/"+obj.Id+"/relationships/"+relationDetails[0],
                            "related": "/"+obj.ObjectType+"/"+obj.Id+"/"+relationDetails[0]
                        },
                        "data": relData
                    };
                  } else {
                      // belongsTo relationship
                      objout.relationships[relationDetails[0]] = {
                        "links": {
                            "self": "/"+obj.ObjectType+"/"+obj.Id+"/relationships/"+relationDetails[0],
                            "related": "/"+obj.ObjectType+"/"+obj.Id+"/"+relationDetails[0]
                        },
                        "data": {"type": relationDetails[1], "id": Number(obj[attr])}
                    };
                  }
              }         
          } else {
              objout[(attr==='ObjectType')?'type':'id'] = obj[attr];
          }
      }
      console.log('Return object is '+JSON.stringify(objout));
      return objout;
  }
  
  const handlingData = (data) => {
      if (Array.isArray(data)) {
          let outdata = [];
          for (let i=0;i<data.length;i++){
              outdata.push(createObject(data[i]));
           }
           return outdata;
      } else {
          return createObject(data);
      }
  }

  const handlingRelationships = (data) => {
    if (Array.isArray(data)) {
        console.warn("Not handling correctly relationship when retrieving list of objects");
    } else {

    }
    return {};
  }
  
  const createResponse = (statusCode, body) => {
      console.log("Body is "+JSON.stringify(body));
      return {
          'statusCode': statusCode,
          'data': handlingData(body),
          'relationships': handlingRelationships(body)
      }
  };
  
  const getMethod = (event, context, callback) => {
  
      let params = {
          TableName: tableName,
      },
      type = event.params.path.type,
      id = Number(event.params.path.id),
      dbGet = {};
  
      if (id) {
          params.Key = {
              'ObjectType': type,
              'Id': id
          };
          dbGet = (params) => { return dynamo.get(params).promise() };
          console.log('EmberDataServerless lambda GET single value with params: ', params);
  
      } else {
          params.KeyConditionExpression = 'ObjectType = :objectType';
          params.ExpressionAttributeValues = { ':objectType': type };
          params.Limit = MAX_OBJECTS;
          dbGet = (params) => { return dynamo.query(params).promise() };
          console.log('EmberDataServerless lambda GET multiple values with params: ', params);
      }
  
      dbGet(params).then( (data) => {
          console.log('EmberDataServerless lambda GET data received: ', data);
  
          if (id && !data.Item) {
              callback(null, createResponse(404, "ITEM NOT FOUND"));
              return;
          } else if (id && data.Item) {
              console.log(`RETRIEVED ITEM SUCCESSFULLY WITH doc = ${data.Item}`);
              callback(null, createResponse(200, data.Item));
          } else {
              console.log('SCANNING TABLE');
              console.log(`RETRIEVED ITEMS SUCCESSFULLY WITH doc = ${data.Items}`);
              callback(null, createResponse(200, data.Items));
          }
          
      }).catch( (err) => { 
          console.log(`GET ITEM FAILED FOR Entry = ${params}, WITH ERROR: ${err}`);
          callback(null, createResponse(500, err));
      });
  };
  
  const putMethod = (event, context, callback) => {
      const body=event['body-json'];
      const attrs = body.data.attributes;
      const relations = body.data.relationships;
      // Without any body content, there is nothing to put...
      if (!body || !attrs) {
          callback(null, createResponse(500, 'No content found in body'));
          return;
      }
      // Retrieving the type and generating a new id for created item
      let type = event.params.path.type,
          id = generateRowId(1);
      // Final content contains at least these two fields + the atributes
      let content = {
          "ObjectType": type,
          "Id": id
      };
      // Adding attributes as column in dynamoDb
      for (var prop in attrs) {
          if (attrs[prop] != null && attrs[prop].length>0)
            // Empty field cannot be persisted
             content[prop] = attrs[prop];
      }
      // Dealing with relationships if any
      if (relations){
          for (var relName in relations) {
            let newCol = relName+'_'+type+'_id';
            let relData = relations[relName]["data"];
            if (!Array.isArray(relData)) {
              content[newCol] = relData["id"];
            } else {
                let ids = [];
                for (var i=0; i<relData.length;i++){
                    let currentData = relData[i];
                    ids.push(currentData["id"]);
                }
                content[newCol] = '['+ids.join(',')+']';
            }
        }
    }
      
    const entry = {
        TableName: tableName,
        Item: content
    };
      console.log('Try saving entity of type '+type+' and content '+JSON.stringify(entry));
      //let dbPut = (entry) => { return dynamo.put(entry).promise() };
      dynamo.put(entry, function(err, data) {
          if (err) {
              console.log("Error", err);
              callback(null, createResponse(500, 'Error '+err));
          } else {
              body.data.id = id;
              body['statusCode'] = 200;
              console.log(`PUT ITEM SUCCEEDED WITH data=`+JSON.stringify(body));
              callback(null, body);
          }
      });
  };
  
  const deleteMethod = (event, context, callback) => {
      let type = event.params.path.type,
          id = Number(event.params.path.id),
          params = {
              'TableName': tableName,
              'Key': {
                  'ObjectType': type,
                  'Id': id
              },
              'ReturnValues': 'ALL_OLD'
          };
      const body=event['body-json'];
  
      let dbDelete = (params) => { return dynamo.delete(params).promise() };
      dbDelete(params).then( (data) => {
          if (!data.Attributes) {
              callback(null, createResponse(404, "ITEM NOT FOUND FOR DELETION"));
              return;
          }
          console.log(`DELETED ITEM OF TYPE ${type} SUCCESSFULLY WITH id = ${id}`);
          callback(null, body);
      }).catch( (err) => { 
          console.log(`DELETE ITEM OF TYPE ${type} FAILED FOR id = ${id}, WITH ERROR: ${err}`);
          callback(null, createResponse(500, err));
      });
  };
  

const callback = (evt, msg) => {console.log(msg);};

const msgGet = {
    "params": {
        "path": {
            "type": "authorities"
        },
        "querystring": {},
        "header": {}
    },
    "context": {
        "httpMethod": "GET",
        "account-id": "995466093880",
        "api-id": "9o6yk638sk",
        "api-key": "test-invoke-api-key",
        "authorizer-principal-id": "",
        "caller": "995466093880",
        "cognito-authentication-provider": "",
        "cognito-authentication-type": "",
        "cognito-identity-id": "",
        "cognito-identity-pool-id": "",
        "stage": "test-invoke-stage",
        "source-ip": "test-invoke-source-ip",
        "user": "995466093880",
        "user-agent": "aws-internal/3 aws-sdk-java/1.11.447 Linux/4.9.124-0.1.ac.198.71.329.metal1.x86_64 OpenJDK_64-Bit_Server_VM/25.192-b12 java/1.8.0_192",
        "user-arn": "arn:aws:iam::995466093880:root",
        "request-id": "b2b534a0-ef42-11e8-bbf2-b564234403bf",
        "resource-id": "anwfak",
        "resource-path": "/EmberDataServerless/{type}"
    }
};

const msgPost = {
    "body-json": {
        "data": {
            "attributes": {
                "login": "freebox",
                "password-hash": "sgsUIHSFDK",
                "first-name": "Laure",
                "last-name": "PINEL",
                "email": "laurepinel@laposte.net",
                "activated": 0,
                "created-by": "Laure PINEL",
                "created-date": "2018-12-26T21:17:32.332Z",
                "last-modified-date": "2018-12-26T21:17:32.332Z"
            },
            "relationships": {
                "authorities": {
                    "data": [
                        {
                            "type": "authorities",
                            "id": "8018931201376941"
                        },
                        {
                            "type": "authorities",
                            "id": "8053142279390032"
                        }
                    ]
                }
            },
            "type": "users"
        }
    },
    "params": {
        "path": {
            "type": "users"
        },
        "querystring": {},
        "header": {
            "Accept": "application/vnd.api+json",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7",
            "cache-control": "no-cache",
            "CloudFront-Forwarded-Proto": "https",
            "CloudFront-Is-Desktop-Viewer": "true",
            "CloudFront-Is-Mobile-Viewer": "false",
            "CloudFront-Is-SmartTV-Viewer": "false",
            "CloudFront-Is-Tablet-Viewer": "false",
            "CloudFront-Viewer-Country": "FR",
            "content-type": "application/json; charset=UTF-8",
            "Cookie": "G_AUTHUSER_H=0; io=INjGzOkni2CdTpLCAAAA; remember-me=QzhwSTVZZmpkTDFxaHJYSkJ5cUs6Znl0MGlnRk1YTGhGbnAySGpEbzE",
            "Host": "dcftkivcqe.execute-api.us-east-1.amazonaws.com",
            "origin": "http://localhost:4200",
            "pragma": "no-cache",
            "Referer": "http://localhost:4200/entity-factory/user?login=freebox&passwordHash=sgsgq+%22%27a%28fqgg&firstName=Laure&lastName=PINEL&email=laurepinel%40laposte.net&imageUrl=&activated=0&createdBy=Laure+PINEL&lastModifiedBy=&authorities=GUEST&authorities=GATSBY",
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/71.0.3578.98 Safari/537.36",
            "Via": "1.1 cbf986a33f2676c4c9c2ef12cabb1a32.cloudfront.net (CloudFront)",
            "X-Amz-Cf-Id": "Z5P_b1_CJqTWY6ZCy1BTPS2E0SBU_uJn3NHjGTj9DfCTgC0Ps_oiQw==",
            "X-Amzn-Trace-Id": "Root=1-5c23f0b0-94250540b6531d00e8bf9600",
            "x-broccoli": "[object Object]",
            "X-Forwarded-For": "::1, 78.119.97.214, 52.46.15.94",
            "x-forwarded-host": "localhost:4200",
            "X-Forwarded-Port": "443",
            "X-Forwarded-Proto": "https",
            "x-requested-with": "XMLHttpRequest"
        }
    },
    "context": {
        "httpMethod": "POST",
        "account-id": "",
        "api-id": "dcftkivcqe",
        "api-key": "",
        "authorizer-principal-id": "",
        "caller": "",
        "cognito-authentication-provider": "",
        "cognito-authentication-type": "",
        "cognito-identity-id": "",
        "cognito-identity-pool-id": "",
        "stage": "staging",
        "source-ip": "78.119.97.214",
        "user": "",
        "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/71.0.3578.98 Safari/537.36",
        "user-arn": "",
        "request-id": "1d5b3483-0954-11e9-b697-131cc1d237db",
        "resource-id": "cmo3pj",
        "resource-path": "/{type}"
    }
};

getMethod(msgGet, {}, callback);
//putMethod(msgPost, {}, callback);